// OpenGLControl.cpp : implementation file
//

#include "stdafx.h"
#include "ShapesDemo.h"
#include "OpenGLControl.h"


// COpenGLControl

IMPLEMENT_DYNAMIC(COpenGLControl, CWnd)

// You will add stuff here!!!!
const char* const COpenGLControl::_ErrorStrings[]= {
    {"No Error"},                     // 0
    {"Unable to get a DC"},           // 1
    {"ChoosePixelFormat failed"},     // 2
    {"SelectPixelFormat failed"},     // 3
    {"wglCreateContext failed"},      // 4
    {"wglMakeCurrent failed"},        // 5
    {"wglDeleteContext failed"},      // 6
    {"SwapBuffers failed"},           // 7
};

COpenGLControl::COpenGLControl()
   // You will add the following line !!!
    : m_hRC(0), m_pDC(0), m_ErrorString(_ErrorStrings[0]), render(false),
	  drawPoints(false), drawLines(false), movePoint(false), culling(false)// Call constructors
{
	pointListEnd = pointList = NULL;
}

void COpenGLControl::Create(CRect rect, CWnd *parent)
{
	// You will add stuff here !!!
  //Register window class
   CString className = AfxRegisterWndClass(
      CS_HREDRAW | CS_VREDRAW | CS_OWNDC,
      NULL,
      (HBRUSH)GetStockObject(BLACK_BRUSH),
      NULL);
 
	//Create the window
   CreateEx(
      0,
      className,
      (CString)"OpenGL",
      WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
      rect,
      parent,
      0);
}

COpenGLControl::~COpenGLControl()
{
	delete pointList;
}


BEGIN_MESSAGE_MAP(COpenGLControl, CWnd)
	ON_WM_CREATE()
	ON_WM_SIZE()
	ON_WM_DESTROY()
	ON_WM_ERASEBKGND()
	ON_WM_PAINT()
	ON_WM_RBUTTONUP()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
END_MESSAGE_MAP()



// COpenGLControl message handlers



int COpenGLControl::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CWnd::OnCreate(lpCreateStruct) == -1)
		return -1;

	// TODO:  Add your specialized creation code here
    // You will add stuff here!!!
     InitializeOpenGL();

	return 0;
}

void COpenGLControl::OnSize(UINT nType, int cx, int cy)
{
	CWnd::OnSize(nType, cx, cy);

	// TODO: Add your message handler code here
    // You are to add a lot of stuff here !!!!
     GLdouble aspect_ratio; // width/height ratio

    if ( 0 >= cx || 0 >= cy )
    {
        return;
    }

	width = (float)cx;
	height = (float)cy;

    SetupViewport( cx, cy );

    // select the projection matrix and clear it
    ::glMatrixMode( GL_PROJECTION );
    ::glLoadIdentity();

    // compute the aspect ratio
    // this will keep all dimension scales equal
    aspect_ratio = (GLdouble)cx/(GLdouble)cy;
   
    // select the viewing volumn
    SetupViewingFrustum( aspect_ratio, cx, cy );

    // switch back to the modelview matrix
    ::glMatrixMode( GL_MODELVIEW );
    ::glLoadIdentity();

    // now perform any viewing transformations
    SetupViewingTransform();
}


void COpenGLControl::OnDestroy()
{
	CWnd::OnDestroy();

	// TODO: Add your message handler code here
    // You will add some stuff here!!!!
    if ( FALSE == ::wglDeleteContext( m_hRC ) )
    {
        SetError(6);
    }

    if ( m_pDC )
    {
        delete m_pDC;
    }


}


BOOL COpenGLControl::OnEraseBkgnd(CDC* pDC)
{
	// TODO: Add your message handler code here and/or call default

	//return CWnd::OnEraseBkgnd(pDC);
	return true;
}


/////////////////////////////////////////////////////////////////////////////
// GL helper functions
// You are to add all new functions here !!!!!!!!!!
// They include:
//    void COpenGLControl::SetError( int e )
//    BOOL COpenGLControl::InitializeOpenGL()
//    BOOL COpenGLControl::SetupPixelFormat()
//    BOOL COpenGLControl::SetupViewport( int cx, int cy )
//    BOOL COpenGLControl::SetupViewingFrustum( GLdouble aspect_ratio )
//    BOOL COpenGLControl::SetupViewingTransform()
//    BOOL COpenGLControl::RenderScene()
//    void COpenGLControl::RenderStockScene()

void COpenGLControl::SetError( int e )
{
    // if there was no previous error,
    // then save this one
    if ( _ErrorStrings[0] == m_ErrorString )
    {
        m_ErrorString = _ErrorStrings[e];
    }
}

BOOL COpenGLControl::InitializeOpenGL()
{
    // Can we put this in the constructor?
    m_pDC = new CClientDC(this);

    if ( NULL == m_pDC ) // failure to get DC
    {
        SetError(1);
        return FALSE;
    }

    if (!SetupPixelFormat())
    {
        return FALSE;
    }

    //n = ::GetPixelFormat(m_pDC->GetSafeHdc());
    //::DescribePixelFormat(m_pDC->GetSafeHdc(), n, sizeof(pfd), &pfd);

    //  CreateRGBPalette();

    if ( 0 == (m_hRC = ::wglCreateContext( m_pDC->GetSafeHdc() ) ) )
    {
        SetError(4);
        return FALSE;
    }

    if ( FALSE == ::wglMakeCurrent( m_pDC->GetSafeHdc(), m_hRC ) )
    {
        SetError(5);
        return FALSE;
    }

    // specify black as clear color
    ::glClearColor( 1.0f, 1.0f, 1.0f, 1.0f );
    // specify the back of the buffer as clear depth
    ::glClearDepth( 1.0f );
    // enable depth testing
    //::glEnable( GL_DEPTH_TEST );

	glPointSize(5);
	glEnable (GL_LINE_SMOOTH);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint (GL_LINE_SMOOTH_HINT, GL_DONT_CARE);

    sceneRotate = 20.0;
	
	drawMode = GL_QUADS;

    return TRUE;
}

BOOL COpenGLControl::SetupPixelFormat()
{
    static PIXELFORMATDESCRIPTOR pfd =
       {
 	 sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
 	 1,                              // version number
 	 PFD_DRAW_TO_WINDOW |            // support window
 	 PFD_SUPPORT_OPENGL |          // support OpenGL
 	 PFD_DOUBLEBUFFER,             // double buffered
 	 PFD_TYPE_RGBA,                  // RGBA type
 	 24,                             // 24-bit color depth
 	 0, 0, 0, 0, 0, 0,               // color bits ignored
 	 0,                              // no alpha buffer
 	 0,                              // shift bit ignored
 	 0,                              // no accumulation buffer
 	 0, 0, 0, 0,                     // accum bits ignored
 	 // 32,                          // 32-bit z-buffer
 	 16,    // NOTE: better performance with 16-bit z-buffer
 	 0,                              // no stencil buffer
 	 0,                              // no auxiliary buffer
 	 PFD_MAIN_PLANE,                 // main layer
 	 0,                              // reserved
 	 0, 0, 0                         // layer masks ignored
       };
    int pixelformat;
 
    if ( 0 == (pixelformat = ::ChoosePixelFormat(m_pDC->GetSafeHdc(), &pfd)) )
    {
        SetError(2);
        return FALSE;
    }

    if ( FALSE == ::SetPixelFormat(m_pDC->GetSafeHdc(), pixelformat, &pfd) )
    {
        SetError(3);
        return FALSE;
    }

    return TRUE;
}

BOOL COpenGLControl::SetupViewport( int cx, int cy )
{
    // select the full client area
	::glViewport(0, 0, cx, cy);
 
    return TRUE;
}

BOOL COpenGLControl::SetupViewingFrustum( GLdouble aspect_ratio, int cx, int cy )
{
    // select a default viewing volumn
    //::gluPerspective( 40.0f, aspect_ratio, .1f, 20.0f );
	::gluOrtho2D(0,cx,cy,0);
    return TRUE;
}


BOOL COpenGLControl::SetupViewingTransform()
{
    // select a default viewing transformation
    // of a 20 degree rotation about the X axis
    // then a -5 unit transformation along Z
    //::glTranslatef( 0.0f, 0.0f, -5.0f );
    //::glRotatef( sceneRotate, 1.0f, 0.0f, 0.0f );
    return TRUE;
}


BOOL COpenGLControl::RenderScene()
{
    return TRUE;
}

// Draw a square surface that looks like a
// black and white checkerboard
void COpenGLControl::RenderStockScene()
{
	CPointList *cur = pointList;

	if (render)
	{
		glColor3f(1.0,0.0,1.0);


		glLineWidth(4);
		glBegin(drawMode);
			while(cur != NULL)
			{
				glVertex2d(cur->x, cur->y);
				cur = cur->next;
			}
		glEnd();

		::glLineWidth(2);
		cur = pointList;
		glColor3f(0.0,0.0,0.0);
		if (drawLines)
		{
			glBegin(GL_LINE_STRIP);
				while(cur != NULL)
				{
					glVertex2d(cur->x, cur->y);
					cur = cur->next;
				}
			glEnd();
		}

		cur = pointList;
		glColor3f(0.0,0.0,1.0);
		if (drawPoints)
		{
			glBegin(GL_POINTS);
				while(cur != NULL)
				{
					glVertex2d(cur->x, cur->y);
					cur = cur->next;
				}
			glEnd();
		}
	}
}

void COpenGLControl::OnPaint()
{
	CPaintDC dc(this); // device context for painting
	// TODO: Add your message handler code here
	// Do not call CWnd::OnPaint() for painting messages
	    // You are to add a lot stuff here !!!!!!

     ::glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    PreRenderScene();

    ::glPushMatrix();
    RenderStockScene();
    ::glPopMatrix();

    ::glFinish();

    if ( FALSE == ::SwapBuffers( m_pDC->GetSafeHdc() ) )
    {
        SetError(7);
    }

}

void COpenGLControl::setDrawMode(GLint mode)
{
	drawMode = mode;
}

void COpenGLControl::OnRButtonUp(UINT nFlags, CPoint point)
{
	// TODO: Add your message handler code here and/or call default
	render = true;

	if(pointListEnd != NULL)
	{
		pointListEnd->next = new CPointList(point);
		pointListEnd = pointListEnd->next;
	}
	else
	{
		pointListEnd = new CPointList(point);
		pointList = pointListEnd;

	}
	RedrawWindow();
}

void COpenGLControl::toggleDrawPoints()
{
	drawPoints?drawPoints=false:drawPoints=true;
	RedrawWindow();
}

void COpenGLControl::toggleDrawLines()
{
	drawLines?drawLines=false:drawLines=true;
	RedrawWindow();
}

void COpenGLControl::drawSelectables()
{
	CPointList *cur = pointList;
	int i = 0;
	glInitNames();
	glPushName(0);
		while(cur != NULL)
		{
			glLoadName(i);
			glBegin(GL_QUADS);
			glVertex2d(cur->x-2, cur->y-2);
			glVertex2d(cur->x+2, cur->y-2);
			glVertex2d(cur->x+2, cur->y+2);
			glVertex2d(cur->x-2, cur->y+2);
			glEnd();
			cur = cur->next;
			i++;
		}
}


void COpenGLControl::OnLButtonDown(UINT nFlags, CPoint point)
{
	// TODO: Add your message handler code here and/or call default
	GLint hits;
	GLuint hitPoints[100];

	glSelectBuffer(100,hitPoints);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	GLint viewport[] = {0,0,0,0};
	glGetIntegerv(GL_VIEWPORT, viewport);
	/*  create 2x2 pixel picking region near cursor location */
	gluOrtho2D(point.x-1,point.x+1,point.y-1,point.y+1);
	
	glMatrixMode (GL_MODELVIEW);
	glRenderMode(GL_SELECT);
	drawSelectables();
	glMatrixMode (GL_PROJECTION);
	glPopMatrix();
	glFlush();
	
	hits = glRenderMode(GL_RENDER);
	

	if(hits > 0)
	{
		movePoint = true;
		activePoint = pointList;

		for (GLuint i = 0; i < hitPoints[3]; i++)
			activePoint = activePoint->next;
	}

}

void COpenGLControl::OnLButtonUp(UINT nFlags, CPoint point)
{
	// TODO: Add your message handler code here and/or call default
	movePoint = false;
}

void COpenGLControl::OnMouseMove(UINT nFlags, CPoint point)
{
	// TODO: Add your message handler code here and/or call default
	if(movePoint)
	{
		activePoint->x = point.x;
		activePoint->y = point.y;
	}
	RedrawWindow();
}

CString COpenGLControl::getPoints()
{
	stringstream buildRet;
	CString ret;
	CPointList *cur = pointList;
	char temp[500];

	while(cur != NULL)
	{
		buildRet << cur->x/width << ", " << cur->y/height;
		cur = cur->next;
		buildRet.getline(temp, 500);
		ret += temp;
		ret += "\r\n";
		buildRet.clear();
	}
	return ret;
}

void COpenGLControl::setPolyType(GLenum mode)
{
    glPolygonMode(GL_FRONT_AND_BACK, mode);
    RedrawWindow();
}

void COpenGLControl::setFrontFace(GLenum direction)
{
    glFrontFace(direction);
    RedrawWindow();

}

void COpenGLControl::toggleCulling()
{
    if(culling)
    {
	glDisable(GL_CULL_FACE);
    }
    else
    {
	glEnable(GL_CULL_FACE);
    }
    culling = !culling;
    RedrawWindow();
}